from __future__ import absolute_import, unicode_literals


class WorkerException(Exception):
    pass


class InvalidMessageFormatException(WorkerException):
    def __init__(self, msg, caught):
        # type: (unicode, Exception) -> None
        super(InvalidMessageFormatException, self).__init__()
        self.msg = msg
        self.caught = caught


class ExecutionFailedException(WorkerException):
    def __init__(self, task_name, caught):
        # type: (unicode, Exception) -> None
        super(ExecutionFailedException, self).__init__()
        self.task_name = task_name
        self.caught = caught


class MaxRetriesReachedException(WorkerException):
    def __init__(self, retries):
        # type: (int) -> None
        super(MaxRetriesReachedException, self).__init__()
        self.retries = retries


class QueueException(WorkerException):
        pass


class InvalidQueueException(QueueException):
    def __init__(self, queue_name):
        # type: (unicode) -> None
        super(InvalidQueueException, self).__init__()
        self.queue_name = queue_name



from __future__ import absolute_import, unicode_literals

from unittest import TestCase

from mock import Mock

from eb_sqs import settings
from eb_sqs.decorators import task
from eb_sqs.worker.queue_client import QueueClient
from eb_sqs.worker.worker import Worker
from eb_sqs.worker.worker_exceptions import MaxRetriesReachedException
from eb_sqs.worker.worker_factory import WorkerFactory


class TestException(Exception):
    pass


@task()
def dummy_task(msg):
    return msg


@task(max_retries=100)
def retries_task(num_of_retries):
    if retries_task.retry_num < num_of_retries:
        retries_task.retry(execute_inline=True)


@task(max_retries=5)
def max_retries_task():
    max_retries_task.retry(execute_inline=True)


@task(max_retries=100)
def repeating_group_task(num_of_retries):
    if repeating_group_task.retry_num < num_of_retries:
        repeating_group_task.retry(execute_inline=True)


@task()
def exception_group_task():
    raise TestException()


@task(max_retries=100)
def exception_repeating_group_task(num_of_retries):
    if exception_repeating_group_task.retry_num == num_of_retries:
        raise TestException()
    else:
        exception_repeating_group_task.retry(execute_inline=True)


@task(max_retries=5)
def max_retries_group_task():
    max_retries_group_task.retry(execute_inline=True)


global_group_mock = Mock()


class WorkerTest(TestCase):
    def setUp(self):
        settings.DEAD_LETTER_MODE = False

        self.queue_mock = Mock(autospec=QueueClient)
        self.worker = Worker(self.queue_mock)

        factory_mock = Mock(autospec=WorkerFactory)
        factory_mock.create.return_value = self.worker
        settings.WORKER_FACTORY = factory_mock

    def test_worker_execution(self):
        msg = '{"id": "id-1", "retry": 0, "queue": "default", "maxRetries": 5, "args": [], "func": "eb_sqs.tests.worker.tests_worker.dummy_task", "kwargs": {"msg": "Hello World!"}}'

        result = self.worker.execute(msg)

        self.assertEqual(result, 'Hello World!')

    def test_worker_execution_dead_letter_queue(self):
        settings.DEAD_LETTER_MODE = True

        msg = '{"id": "id-1", "groupId": "group-5", "retry": 0, "queue": "default", "maxRetries": 5, "args": [], "func": "eb_sqs.tests.worker.tests_worker.dummy_task", "kwargs": {"msg": "Hello World!"}}'

        result = self.worker.execute(msg)

        self.assertIsNone(result)

    def test_delay(self):
        self.worker.delay(None, 'queue', dummy_task, [], {'msg': 'Hello World!'}, 5, False, 3, False)

        self.queue_mock.add_message.assert_called_once()
        queue_delay = self.queue_mock.add_message.call_args[0][2]
        self.assertEqual(queue_delay, 3)

    def test_delay_inline(self):
        result = self.worker.delay(None, 'queue', dummy_task, [], {'msg': 'Hello World!'}, 5, False, 0, True)

        self.queue_mock.add_message.assert_not_called()
        self.assertEqual(result, 'Hello World!')

    def test_retry_max_reached_execution(self):
        with self.assertRaises(MaxRetriesReachedException):
            max_retries_task.delay(execute_inline=True)

    def test_retry_no_limit(self):
        retries_task.delay(10, execute_inline=True)

        self.assertEqual(retries_task.retry_num, 10)



SECRET_KEY = "secret-test-key"

INSTALLED_APPS = (
    'eb_sqs',
)

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': ':memory:',
    }
}

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [
            # insert your TEMPLATE_DIRS here
        ],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                # Insert your TEMPLATE_CONTEXT_PROCESSORS here or use this
                # list if you haven't customized them:
                'django.contrib.auth.context_processors.auth',
                'django.template.context_processors.debug',
                'django.template.context_processors.i18n',
                'django.template.context_processors.media',
                'django.template.context_processors.static',
                'django.template.context_processors.tz',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

ROOT_URLCONF = 'eb_sqs.urls'

LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'handlers': {
        'console': {
            'level': 'DEBUG',
            'class': 'logging.StreamHandler'
        },
    },
    'loggers': {
        'eb_sqs': {
            'handlers': ['console'],
            'level': 'CRITICAL'
        },
    }
}



from __future__ import absolute_import, unicode_literals

from abc import ABCMeta, abstractmethod


class QueueClientException(Exception):
    pass


class QueueDoesNotExistException(QueueClientException):
    def __init__(self, queue_name):
        # type: (unicode) -> None
        super(QueueDoesNotExistException, self).__init__()
        self.queue_name = queue_name


class QueueClient(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def add_message(self, queue_name, msg, delay):
        # type: (unicode, unicode, int) -> None
        pass



from __future__ import absolute_import, unicode_literals

import json
from unittest import TestCase

from eb_sqs.worker.worker_task import WorkerTask


class TestObject(object):
    def __init__(self):
        # type: () -> None
        super(TestObject, self).__init__()
        self.message = 'Test'


def dummy_function():
    pass


class WorkerTaskTest(TestCase):
    def setUp(self):
        self.dummy_msg = '{"queue": "default", "retryId": "retry-uuid", "retry": 0, "func": "eb_sqs.tests.worker.tests_worker_task.dummy_function", "kwargs": {}, "maxRetries": 5, "args": [], "pickle": false, "id": "id-1", "groupId": "group-5"}'

    def test_serialize_worker_task(self):
        worker_task = WorkerTask('id-1', 'group-5', 'default', dummy_function, [], {}, 5, 0, 'retry-uuid', False)
        msg = worker_task.serialize()

        self.assertDictEqual(json.loads(msg), json.loads(self.dummy_msg))

    def test_deserialize_worker_task(self):
        worker_task = WorkerTask.deserialize(self.dummy_msg)

        self.assertEqual(worker_task.id, 'id-1')
        self.assertEqual(worker_task.group_id, 'group-5')
        self.assertEqual(worker_task.queue, 'default')
        self.assertEqual(worker_task.func, dummy_function)
        self.assertEqual(worker_task.args, [])
        self.assertEqual(worker_task.kwargs, {})
        self.assertEqual(worker_task.max_retries, 5)
        self.assertEqual(worker_task.retry, 0)
        self.assertEqual(worker_task.retry_id, 'retry-uuid')

    def test_serialize_pickle(self):
        worker_task1 = WorkerTask('id-1', None, 'default', dummy_function, [], {'object': TestObject()}, 5, 0, None, True)
        msg = worker_task1.serialize()

        worker_task2 = WorkerTask.deserialize(msg)
        self.assertEqual(worker_task2.args, worker_task1.args)
        self.assertEqual(worker_task2.kwargs['object'].message, worker_task1.kwargs['object'].message)



from __future__ import absolute_import, unicode_literals

import boto3
from botocore.config import Config
from botocore.exceptions import ClientError

from eb_sqs import settings
from eb_sqs.worker.queue_client import QueueClient, QueueDoesNotExistException, QueueClientException


class SqsQueueClient(QueueClient):
    def __init__(self):
        # type: () -> None
        self.sqs = boto3.resource('sqs',
                                  region_name=settings.AWS_REGION,
                                  config=Config(retries={'max_attempts': settings.AWS_MAX_RETRIES})
                                  )
        self.queue_cache = {}

    def _get_queue(self, queue_name, use_cache=True):
        # type: (unicode, bool) -> Any
        full_queue_name = '{}{}'.format(settings.QUEUE_PREFIX, queue_name)

        queue = self._get_sqs_queue(full_queue_name, use_cache)
        if not queue:
            queue = self._add_sqs_queue(full_queue_name)

        return queue

    def _get_sqs_queue(self, queue_name, use_cache):
        # type: (unicode, bool) -> Any
        if use_cache and self.queue_cache.get(queue_name):
            return self.queue_cache[queue_name]

        try:
            queue = self.sqs.get_queue_by_name(QueueName=queue_name)
            self.queue_cache[queue_name] = queue
            return queue
        except ClientError as ex:
            error_code = ex.response.get('Error', {}).get('Code', None)
            if error_code == 'AWS.SimpleQueueService.NonExistentQueue':
                return None
            else:
                raise ex

    def _add_sqs_queue(self, queue_name):
        # type: (unicode) -> Any
        if settings.AUTO_ADD_QUEUE:
            queue = self.sqs.create_queue(
                QueueName=queue_name,
                Attributes={
                    'MessageRetentionPeriod': settings.QUEUE_MESSAGE_RETENTION,
                    'VisibilityTimeout': settings.QUEUE_VISIBILITY_TIMEOUT
                }
            )
            self.queue_cache[queue_name] = queue
            return queue
        else:
            raise QueueDoesNotExistException(queue_name)

    def add_message(self, queue_name, msg, delay):
        # type: (unicode, unicode, int) -> None
        try:
            queue = self._get_queue(queue_name)
            try:
                queue.send_message(
                    MessageBody=msg,
                    DelaySeconds=delay
                )
            except ClientError as ex:
                if ex.response.get('Error', {}).get('Code', None) == 'AWS.SimpleQueueService.NonExistentQueue':
                    queue = self._get_queue(queue_name, use_cache=False)
                    queue.send_message(
                        MessageBody=msg,
                        DelaySeconds=delay
                    )
                else:
                    raise ex
        except QueueDoesNotExistException:
            raise
        except Exception as ex:
            raise QueueClientException(ex)



from __future__ import absolute_import, unicode_literals

from django.core.management import BaseCommand, CommandError

from eb_sqs.worker.service import WorkerService


class Command(BaseCommand):
    help = 'Command to process tasks from one or more SQS queues'

    def add_arguments(self, parser):
        parser.add_argument('--queues', '-q',
                            dest='queue_names',
                            help='Name of queues to process, separated by commas')

    def handle(self, *args, **options):
        if not options['queue_names']:
            raise CommandError('Queue names (--queues) not specified')

        queue_names = [queue_name.rstrip() for queue_name in options['queue_names'].split(',')]

        WorkerService().process_queues(queue_names)



from __future__ import absolute_import, unicode_literals

from unittest import TestCase

from mock import Mock

from eb_sqs import settings
from eb_sqs.decorators import task
from eb_sqs.worker.worker import Worker
from eb_sqs.worker.worker_factory import WorkerFactory


@task()
def dummy_task(msg):
    # type: (unicode) -> None
    if not msg:
        raise Exception('No message')

@task(queue_name='CustomQueue')
def dummy_task_custom_queue():
    # type: (unicode) -> None
    pass

@task()
def dummy_retry_task(msg):
    # type: (unicode) -> None
    if dummy_retry_task.retry_num == 0:
        dummy_retry_task.retry()
    else:
        if not msg:
            raise Exception('No message')


class DecoratorsTest(TestCase):
    def setUp(self):
        self.worker_mock = Mock(autospec=Worker)

        factory_mock = Mock(autospec=WorkerFactory)
        factory_mock.create.return_value = self.worker_mock
        settings.WORKER_FACTORY = factory_mock

    def test_delay_decorator(self):
        dummy_task.delay('Hello World!')
        self.worker_mock.delay.assert_called_once()

    def test_delay_custom_queue_decorator(self):
        dummy_task_custom_queue.delay()

        call_args = self.worker_mock.delay.call_args
        self.assertTrue('CustomQueue' in call_args[0])

    def test_delay_custom_queue_as_param_decorator(self):
        dummy_task_custom_queue.delay(queue_name='OtherQueue')

        call_args = self.worker_mock.delay.call_args
        self.assertTrue('OtherQueue' in call_args[0])

    def test_delay_decorator_parameters(self):
        dummy_task.delay('Hello World!', group_id='group-5', delay=5, execute_inline=True)

        self.worker_mock.delay.assert_called_once()

        # Parameter should have been removed from kwargs before being passed to the actual function
        kwargs = self.worker_mock.delay.call_args[0][4]
        self.assertEqual(kwargs, {})

    def test_retry_decorator(self):
        dummy_retry_task.delay('Hello World!')
        self.worker_mock.delay.assert_called_once()



from unittest import TestCase

from mock import Mock, call

from eb_sqs import settings
from eb_sqs.auto_tasks.exceptions import RetryableTaskException
from eb_sqs.auto_tasks.service import AutoTaskService, _auto_task_wrapper


class TestService:
    TEST_MOCK = None
    MAX_RETRY_NUM = 5

    def __init__(self, auto_task_service=None):
        self._auto_task_service = auto_task_service or AutoTaskService()

        self._auto_task_service.register_task(self.task_method)
        self._auto_task_service.register_task(self.task_retry_method, max_retries=self.MAX_RETRY_NUM)

        self._auto_task_service.register_task(self.task_recursive_method)
        self._auto_task_service.register_task(self.task_other_method)

    def task_method(self, *args, **kwargs):
        self.TEST_MOCK.task_method(*args, **kwargs)

    def task_retry_method(self, *args, **kwargs):
        self.TEST_MOCK.task_retry_method(*args, **kwargs)

        def max_retry_fun():
            self.TEST_MOCK.task_max_retry_method(*args, **kwargs)

        raise RetryableTaskException(Exception('Test'), max_retries_func=max_retry_fun)

    def non_task_method(self):
        self.TEST_MOCK.non_task_method()

    def task_recursive_method(self, tries=2):
        if tries > 0:
            self.task_recursive_method(tries=tries - 1)
        else:
            self.task_other_method()

    def task_other_method(self):
        self.TEST_MOCK.task_other_method()


class AutoTasksTest(TestCase):
    def setUp(self):
        self._test_service = TestService()

        self._args = [5, '6']
        self._kwargs = {'p1': 'bla', 'p2': 130}

        settings.EXECUTE_INLINE = True

        TestService.TEST_MOCK = Mock()

    def test_task_method(self):
        self._test_service.task_method(*self._args, **self._kwargs)

        TestService.TEST_MOCK.task_method.assert_called_once_with(*self._args, **self._kwargs)

    def test_task_retry_method(self):
        self._test_service.task_retry_method(*self._args, **self._kwargs)

        TestService.TEST_MOCK.task_retry_method.assert_has_calls([call(*self._args, **self._kwargs)] * TestService.MAX_RETRY_NUM)

        TestService.TEST_MOCK.task_max_retry_method.assert_called_once_with(*self._args, **self._kwargs)

    def test_non_task_method(self):
        _auto_task_wrapper.delay(
            self._test_service.__class__.__module__,
            self._test_service.__class__.__name__,
            TestService.non_task_method.__name__,
            execute_inline=True
        )

        TestService.TEST_MOCK.non_task_method.assert_not_called()

    def test_task_recursive_method(self):
        self._test_service.task_recursive_method()

        TestService.TEST_MOCK.task_other_method.assert_called_once_with()



from __future__ import absolute_import, unicode_literals

from eb_sqs.aws.sqs_queue_client import SqsQueueClient
from eb_sqs.worker.worker import Worker
from eb_sqs.worker.worker_factory import WorkerFactory


class SqsWorkerFactory(WorkerFactory):
    _WORKER = None # type: Worker

    def __init__(self):
        super(SqsWorkerFactory, self).__init__()

    def create(self):
        if not SqsWorkerFactory._WORKER:
            SqsWorkerFactory._WORKER = Worker(SqsQueueClient())
        return SqsWorkerFactory._WORKER



from __future__ import absolute_import, unicode_literals

import time
from unittest import TestCase
from mock import patch

import boto3
from botocore.exceptions import ClientError
from moto import mock_sqs

from eb_sqs import settings
from eb_sqs.aws.sqs_queue_client import SqsQueueClient
from eb_sqs.worker.queue_client import QueueDoesNotExistException


class AwsQueueClientTest(TestCase):
    def setUp(self):
        settings.QUEUE_PREFIX = 'eb-sqs-'

    @mock_sqs()
    def test_add_message(self):
        sqs = boto3.resource('sqs',
                             region_name=settings.AWS_REGION)
        queue = sqs.create_queue(QueueName='eb-sqs-default')

        queue_client = SqsQueueClient()

        queue_client.add_message('default', 'msg', 0)

        queue.reload()
        self.assertEqual(queue.attributes["ApproximateNumberOfMessages"], '1')

    @mock_sqs()
    def test_add_message_delayed(self):
        delay = 1
        sqs = boto3.resource('sqs',
                             region_name=settings.AWS_REGION)
        queue = sqs.create_queue(QueueName='eb-sqs-default')
        queue_client = SqsQueueClient()

        queue_client.add_message('default', 'msg', delay)

        queue.reload()
        self.assertEqual(queue.attributes["ApproximateNumberOfMessages"], '0')

        time.sleep(delay + 0.1)

        queue.reload()
        self.assertEqual(queue.attributes["ApproximateNumberOfMessages"], '1')

    @mock_sqs()
    def test_add_message_wrong_queue(self):
        sqs = boto3.resource('sqs',
                             region_name=settings.AWS_REGION)
        sqs.create_queue(QueueName='default')
        queue_client = SqsQueueClient()

        with self.assertRaises(QueueDoesNotExistException):
            queue_client.add_message('invalid', 'msg', 0)

    @mock_sqs()
    def test_auto_add_queue(self):
        settings.AUTO_ADD_QUEUE = True

        queue_name = 'test-queue'

        sqs = boto3.resource('sqs',
                             region_name=settings.AWS_REGION)

        queue_client = SqsQueueClient()

        queue_client.add_message(queue_name, 'msg', 0)

        full_queue_name = settings.QUEUE_PREFIX + queue_name

        queue = sqs.get_queue_by_name(QueueName=full_queue_name)

        self.assertEqual(queue.attributes["ApproximateNumberOfMessages"], '1')

        queue.delete()

        # moto throws exception inconsistent with boto, thus the patching
        with patch.object(queue_client.queue_cache[full_queue_name], 'send_message') as send_message_fn:
            send_message_fn.side_effect = ClientError({'Error': {'Code': 'AWS.SimpleQueueService.NonExistentQueue'}}, None)

            queue_client.add_message(queue_name, 'msg', 0)

        queue = sqs.get_queue_by_name(QueueName=full_queue_name)

        self.assertEqual(queue.attributes["ApproximateNumberOfMessages"], '1')

        settings.AUTO_ADD_QUEUE = False



from __future__ import absolute_import, unicode_literals

from contextlib import contextmanager

from django.db import reset_queries, close_old_connections


@contextmanager
def django_db_management():
    # type: () -> None
    reset_queries()
    close_old_connections()
    try:
        yield
    finally:
        close_old_connections()



from __future__ import absolute_import, unicode_literals

from abc import ABCMeta, abstractmethod

from eb_sqs import settings
from eb_sqs.worker.worker import Worker


class WorkerFactory(object):
    __metaclass__ = ABCMeta

    def __init__(self):
        # type: () -> None
        super(WorkerFactory, self).__init__()

    @abstractmethod
    def create(self):
        # type: () -> Worker
        pass

    @staticmethod
    def default():
        # type: () -> WorkerFactory
        if not settings.WORKER_FACTORY:
            from eb_sqs.worker.sqs_worker_factory import SqsWorkerFactory
            return SqsWorkerFactory()
        else:
            return settings.WORKER_FACTORY



from __future__ import absolute_import, unicode_literals

from eb_sqs import settings
from eb_sqs.worker.worker_factory import WorkerFactory
from eb_sqs.worker.worker_task import WorkerTask


def _get_kwarg_val(kwargs, key, default):
    # type: (dict, str, Any) -> Any
    return kwargs.pop(key, default) if kwargs else default


def func_delay_decorator(func, queue_name, max_retries_count, use_pickle):
    # type: (Any, str, int, bool) -> (tuple, dict)
    def wrapper(*args, **kwargs):
        # type: (tuple, dict) -> Any
        queue = _get_kwarg_val(kwargs, 'queue_name', queue_name if queue_name else settings.DEFAULT_QUEUE)
        max_retries = _get_kwarg_val(kwargs, 'max_retries', max_retries_count if max_retries_count else settings.DEFAULT_MAX_RETRIES)
        pickle = _get_kwarg_val(kwargs, 'use_pickle', use_pickle if use_pickle else settings.USE_PICKLE)

        execute_inline = _get_kwarg_val(kwargs, 'execute_inline', False) or settings.EXECUTE_INLINE
        delay = _get_kwarg_val(kwargs, 'delay',  settings.DEFAULT_DELAY)
        group_id = _get_kwarg_val(kwargs, 'group_id', None)

        worker = WorkerFactory.default().create()
        return worker.delay(group_id, queue, func, args, kwargs, max_retries, pickle, delay, execute_inline)

    return wrapper


def func_retry_decorator(worker_task):
    # type: (WorkerTask) -> (tuple, dict)
    def wrapper(*args, **kwargs):
        # type: (tuple, dict) -> Any
        execute_inline = _get_kwarg_val(kwargs, 'execute_inline', False) or settings.EXECUTE_INLINE
        delay = _get_kwarg_val(kwargs, 'delay', settings.DEFAULT_DELAY)
        count_retries = _get_kwarg_val(kwargs, 'count_retries', settings.DEFAULT_COUNT_RETRIES)

        worker = WorkerFactory.default().create()
        return worker.retry(worker_task, delay, execute_inline, count_retries)
    return wrapper


class task(object):
    def __init__(self, queue_name=None, max_retries=None, use_pickle=None):
        # type: (str, int, bool) -> None
        self.queue_name = queue_name
        self.max_retries = max_retries
        self.use_pickle = use_pickle

    def __call__(self, *args, **kwargs):
        # type: (tuple, dict) -> Any
        func = args[0]
        func.retry_num = 0
        func.delay = func_delay_decorator(func, self.queue_name, self.max_retries, self.use_pickle)
        return func



from __future__ import absolute_import, unicode_literals

import logging
import uuid

from eb_sqs import settings
from eb_sqs.worker.queue_client import QueueDoesNotExistException, QueueClient, QueueClientException
from eb_sqs.worker.worker_exceptions import InvalidMessageFormatException, ExecutionFailedException, \
    MaxRetriesReachedException, InvalidQueueException, QueueException
from eb_sqs.worker.worker_task import WorkerTask

logger = logging.getLogger("eb_sqs")


class Worker(object):
    def __init__(self, queue_client):
        # type: (QueueClient) -> None
        super(Worker, self).__init__()
        self.queue_client = queue_client

    def execute(self, msg):
        # type: (unicode) -> Any
        try:
            worker_task = WorkerTask.deserialize(msg)
        except Exception as ex:
            logger.exception(
                'Message %s is not a valid worker task: %s',
                msg,
                ex
            )

            raise InvalidMessageFormatException(msg, ex)

        try:
            if settings.DEAD_LETTER_MODE:
                # If in dead letter mode only try to run callback. Do not execute task.
                logger.debug(
                    'Task %s (%s, retry-id: %s) not executed (dead letter queue)',
                    worker_task.abs_func_name,
                    worker_task.id,
                    worker_task.retry_id,
                )

            else:
                logger.debug(
                    'Execute task %s (%s, retry-id: %s) with args: %s and kwargs: %s',
                    worker_task.abs_func_name,
                    worker_task.id,
                    worker_task.retry_id,
                    worker_task.args,
                    worker_task.kwargs
                )

                return self._execute_task(worker_task)
        except QueueException:
            raise
        except MaxRetriesReachedException:
            raise
        except Exception as ex:
            logger.exception(
                'Task %s (%s, retry-id: %s) failed to execute with args: %s and kwargs: %s: %s',
                worker_task.abs_func_name,
                worker_task.id,
                worker_task.retry_id,
                worker_task.args,
                worker_task.kwargs,
                ex
            )

            raise ExecutionFailedException(worker_task.abs_func_name, ex)

    def delay(self, group_id, queue_name, func, args, kwargs, max_retries, use_pickle, delay, execute_inline):
        # type: (unicode, unicode, Any, tuple, dict, int, bool, int, bool) -> Any
        worker_task = WorkerTask(str(uuid.uuid4()), group_id, queue_name, func, args, kwargs, max_retries, 0, None, use_pickle)
        return self._enqueue_task(worker_task, delay, execute_inline, False, True)

    def retry(self, worker_task, delay, execute_inline, count_retries):
        # type: (WorkerTask, int, bool, bool) -> Any
        worker_task = worker_task.copy(settings.FORCE_SERIALIZATION)
        worker_task.retry_id = str(uuid.uuid4())
        return self._enqueue_task(worker_task, delay, execute_inline, True, count_retries)

    def _enqueue_task(self, worker_task, delay, execute_inline, is_retry, count_retries):
        # type: (WorkerTask, int, bool, bool, bool) -> Any
        try:
            if is_retry and count_retries:
                worker_task.retry += 1
                if worker_task.retry >= worker_task.max_retries:
                    raise MaxRetriesReachedException(worker_task.retry)

            logger.debug('%s task %s (%s, retry-id: %s): %s, %s (%s%s)',
                         'Retrying' if is_retry else 'Delaying',
                         worker_task.abs_func_name,
                         worker_task.id,
                         worker_task.retry_id,
                         worker_task.args,
                         worker_task.kwargs,
                         worker_task.queue,
                         ', inline' if execute_inline else '')

            if execute_inline:
                return self._execute_task(worker_task)
            else:
                self.queue_client.add_message(worker_task.queue, worker_task.serialize(), delay)
                return None
        except QueueDoesNotExistException as ex:
            raise InvalidQueueException(ex.queue_name)
        except QueueClientException as ex:
            logger.warning('Task %s (%s, retry-id: %s) failed to enqueue to %s: %s',
                           worker_task.abs_func_name,
                           worker_task.id,
                           worker_task.retry_id,
                           worker_task.queue,
                           ex)

            raise QueueException()

    @classmethod
    def _execute_task(cls, worker_task):
        # type: (WorkerTask) -> Any
        result = worker_task.execute()
        return result



from __future__ import absolute_import, unicode_literals

from django.conf import settings

AWS_REGION = getattr(settings, 'EB_AWS_REGION', 'us-east-1')  # type: str

MAX_NUMBER_OF_MESSAGES = getattr(settings, 'EB_SQS_MAX_NUMBER_OF_MESSAGES', 10)  # type: int
WAIT_TIME_S = getattr(settings, 'EB_SQS_WAIT_TIME_S', 2)  # type: int
NO_QUEUES_WAIT_TIME_S = getattr(settings, 'NO_QUEUES_WAIT_TIME_S', 5)  # type: int

AUTO_ADD_QUEUE = getattr(settings, 'EB_SQS_AUTO_ADD_QUEUE', False)  # type: bool
QUEUE_PREFIX = getattr(settings, 'EB_SQS_QUEUE_PREFIX', '')  # type: str
DEFAULT_QUEUE = getattr(settings, 'EB_SQS_DEFAULT_QUEUE', 'eb-sqs-default')  # type: str

EXECUTE_INLINE = getattr(settings, 'EB_SQS_EXECUTE_INLINE', False)  # type: bool
FORCE_SERIALIZATION = getattr(settings, 'EB_SQS_FORCE_SERIALIZATION', False)  # type: bool

DEFAULT_DELAY = getattr(settings, 'EB_SQS_DEFAULT_DELAY', 0)  # type: int
DEFAULT_MAX_RETRIES = getattr(settings, 'EB_SQS_DEFAULT_MAX_RETRIES', 0)  # type: int
DEFAULT_COUNT_RETRIES = getattr(settings, 'EB_SQS_DEFAULT_COUNT_RETRIES', True)  # type: bool

USE_PICKLE = getattr(settings, 'EB_SQS_USE_PICKLE', False)  # type: bool

WORKER_FACTORY = getattr(settings, 'EB_SQS_WORKER_FACTORY', None)  # type: WorkerFactory

DEAD_LETTER_MODE = getattr(settings, 'EB_SQS_DEAD_LETTER_MODE', False)  # type: bool

AWS_MAX_RETRIES = getattr(settings, 'EB_SQS_AWS_MAX_RETRIES', 30)  # type: int

REFRESH_PREFIX_QUEUES_S = getattr(settings, 'EB_SQS_REFRESH_PREFIX_QUEUES_S', 10)  # type: int

QUEUE_MESSAGE_RETENTION = getattr(settings, 'EB_SQS_QUEUE_MESSAGE_RETENTION', '1209600')  # type: str
QUEUE_VISIBILITY_TIMEOUT = getattr(settings, 'EB_SQS_QUEUE_VISIBILITY_TIMEOUT', '300')  # type: str

MIN_HEALTHCHECK_WRITE_PERIOD_S = getattr(settings, 'EB_SQS_MIN_HEALTHCHECK_WRITE_PERIOD_S', 10)  # type: int
HEALTHCHECK_UNHEALTHY_PERIOD_S = getattr(settings, 'EB_SQS_HEALTHCHECK_UNHEALTHY_PERIOD_S', int(QUEUE_VISIBILITY_TIMEOUT))  # type: int
HEALTHCHECK_FILE_NAME = getattr(settings, 'EB_SQS_HEALTHCHECK_FILE_NAME', 'healthcheck.txt')  # type: str



import sys

import logging
from datetime import timedelta

from django.core.management import BaseCommand
from django.utils import timezone
from django.utils.dateparse import parse_datetime

from eb_sqs import settings

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    help = 'Checks the SQS worker is healthy, and if not returns a failure code'

    def add_arguments(self, parser):
        pass

    def handle(self, *args, **options):
        try:
            with open(settings.HEALTHCHECK_FILE_NAME, 'r') as file:
                last_healthcheck_date_str = file.readlines()[0]

                if parse_datetime(last_healthcheck_date_str) < timezone.now() - timedelta(seconds=settings.HEALTHCHECK_UNHEALTHY_PERIOD_S):
                    self._return_failure()
        except Exception:
            self._return_failure()

    @staticmethod
    def _return_failure():
        logger.warning('[django-eb-sqs] Health check failed')
        sys.exit(1)



class RetryableTaskException(Exception):
    def __init__(self, inner, delay=None, count_retries=None, max_retries_func=None):
        # type: (Exception, int, bool, Any) -> None
        self._inner = inner

        self.delay = delay
        self.count_retries = count_retries
        self.max_retries_func = max_retries_func

    def __repr__(self):
        # type: () -> str
        return repr(self._inner)



import importlib
import logging

from eb_sqs.auto_tasks.exceptions import RetryableTaskException
from eb_sqs.decorators import task
from eb_sqs.worker.worker_exceptions import MaxRetriesReachedException

logger = logging.getLogger(__name__)


@task()
def _auto_task_wrapper(module_name, class_name, func_name, *args, **kwargs):
    try:
        logger.debug(
            'Invoke _auto_task_wrapper with module: %s class: %s func: %s args: %s and kwargs: %s',
            module_name,
            class_name,
            func_name,
            args,
            kwargs
        )

        module = importlib.import_module(module_name)  # import module
        class_ = getattr(module, class_name)  # find class

        auto_task_executor_service = _AutoTaskExecutorService(func_name)
        instance = class_(auto_task_service=auto_task_executor_service)  # instantiate class using _AutoTaskExecutorService

        executor_func_name = auto_task_executor_service.get_executor_func_name()
        if executor_func_name:
            getattr(instance, executor_func_name)(*args, **kwargs)  # invoke method on instance
        else:
            logger.error(
                'Trying to invoke _auto_task_wrapper for unregistered task with module: %s class: %s func: %s args: %s and kwargs: %s',
                module_name,
                class_name,
                func_name,
                args,
                kwargs
            )
    except RetryableTaskException as exc:
        try:
            retry_kwargs = {}

            if exc.delay is not None:
                retry_kwargs['delay'] = exc.delay

            if exc.count_retries is not None:
                retry_kwargs['count_retries'] = exc.count_retries

            _auto_task_wrapper.retry(**retry_kwargs)
        except MaxRetriesReachedException:
            if exc.max_retries_func:
                exc.max_retries_func()
            else:
                # by default log an error
                logger.error('Reached max retries in auto task {}.{}.{} with error: {}'.format(module_name, class_name, func_name, repr(exc)))


class AutoTaskService(object):
    def register_task(self, method, queue_name=None, max_retries=None):
        # type: (Any, str, int) -> None
        instance = method.__self__
        class_ = instance.__class__
        func_name = method.__name__

        def _auto_task_wrapper_invoker(*args, **kwargs):
            if queue_name is not None:
                kwargs['queue_name'] = queue_name

            if max_retries is not None:
                kwargs['max_retries'] = max_retries

            _auto_task_wrapper.delay(
                class_.__module__,
                class_.__name__,
                func_name,
                *args, **kwargs
            )

        setattr(instance, func_name, _auto_task_wrapper_invoker)


class _AutoTaskExecutorService(AutoTaskService):
    def __init__(self, func_name):
        # type: (str) -> None
        self._func_name = func_name

        self._executor_func_name = None

    def register_task(self, method, queue_name=None, max_retries=None):
        # type: (Any, str, int) -> None
        if self._func_name == method.__name__:
            # circuit breaker to allow actually executing the method once
            instance = method.__self__

            self._executor_func_name = self._func_name + '__auto_task_executor__'
            setattr(instance, self._executor_func_name, getattr(instance, self._func_name))

        super(_AutoTaskExecutorService, self).register_task(method, queue_name, max_retries)

    def get_executor_func_name(self):
        # type: () -> str
        return self._executor_func_name



from __future__ import absolute_import, unicode_literals

import base64
import importlib
import json
import uuid

from eb_sqs import settings

try:
    import cPickle as pickle
except Exception:
    import pickle


class WorkerTask(object):
    def __init__(self, id, group_id, queue, func, args, kwargs, max_retries, retry, retry_id, use_pickle):
        # type: (str, unicode, unicode, Any, tuple, dict, int, int, unicode, bool) -> None
        super(WorkerTask, self).__init__()
        self.id = id
        self.group_id = group_id
        self.queue = queue
        self.func = func
        self.args = args
        self.kwargs = kwargs
        self.max_retries = max_retries
        self.retry = retry
        self.retry_id = retry_id
        self.use_pickle = use_pickle

        self.abs_func_name = '{}.{}'.format(self.func.__module__, self.func.__name__)

    def execute(self):
        # type: () -> Any
        from eb_sqs.decorators import func_retry_decorator
        self.func.retry_num = self.retry
        self.func.retry = func_retry_decorator(worker_task=self)
        return self.func(*self.args, **self.kwargs)

    def serialize(self):
        # type: () -> unicode
        args = WorkerTask._pickle_args(self.args) if self.use_pickle else self.args
        kwargs = WorkerTask._pickle_args(self.kwargs) if self.use_pickle else self.kwargs

        task = {
                'id': self.id,
                'groupId': self.group_id,
                'queue': self.queue,
                'func': self.abs_func_name,
                'args': args,
                'kwargs': kwargs,
                'maxRetries': self.max_retries,
                'retry': self.retry,
                'retryId': self.retry_id,
                'pickle': self.use_pickle,
            }

        return json.dumps(task)

    def copy(self, use_serialization):
        # type: (bool) -> WorkerTask
        if use_serialization:
            return WorkerTask.deserialize(self.serialize())
        else:
            return WorkerTask(
                self.id,
                self.group_id,
                self.queue,
                self.func,
                self.args,
                self.kwargs,
                self.max_retries,
                self.retry,
                self.retry_id,
                self.use_pickle,
            )

    @staticmethod
    def _pickle_args(args):
        # type: (Any) -> unicode
        return base64.b64encode(pickle.dumps(args, pickle.HIGHEST_PROTOCOL)).decode('utf-8')

    @staticmethod
    def deserialize(msg):
        # type: (unicode) -> WorkerTask
        task = json.loads(msg)

        id = task.get('id', str(uuid.uuid4()))
        group_id = task.get('groupId')

        abs_func_name = task['func']
        func_name = abs_func_name.split(".")[-1]
        func_path = ".".join(abs_func_name.split(".")[:-1])
        func_module = importlib.import_module(func_path)

        func = getattr(func_module, func_name)

        use_pickle = task.get('pickle', False)
        queue = task.get('queue', settings.DEFAULT_QUEUE)

        task_args = task.get('args', [])
        args = WorkerTask._unpickle_args(task_args) if use_pickle else task_args

        kwargs = WorkerTask._unpickle_args(task['kwargs']) if use_pickle else task['kwargs']

        max_retries = task.get('maxRetries', settings.DEFAULT_MAX_RETRIES)
        retry = task.get('retry', 0)
        retry_id = task.get('retryId')

        return WorkerTask(id, group_id, queue, func, args, kwargs, max_retries, retry, retry_id, use_pickle)

    @staticmethod
    def _unpickle_args(args):
        # type: (unicode) -> dict
        return pickle.loads(base64.b64decode(args.encode('utf-8')))



from __future__ import absolute_import, unicode_literals

import logging
from datetime import timedelta
from time import sleep

import boto3
from botocore.config import Config
from botocore.exceptions import ClientError
import django.dispatch
from django.utils import timezone

from eb_sqs import settings
from eb_sqs.worker.commons import django_db_management
from eb_sqs.worker.worker import Worker
from eb_sqs.worker.worker_exceptions import ExecutionFailedException
from eb_sqs.worker.worker_factory import WorkerFactory

logger = logging.getLogger(__name__)

MESSAGES_RECEIVED = django.dispatch.Signal(providing_args=['messages'])
MESSAGES_PROCESSED = django.dispatch.Signal(providing_args=['messages'])
MESSAGES_DELETED = django.dispatch.Signal(providing_args=['messages'])


class WorkerService(object):
    _PREFIX_STR = 'prefix:'
    _RECEIVE_COUNT_ATTRIBUTE = 'ApproximateReceiveCount'

    def process_queues(self, queue_names):
        # type: (list) -> None
        logger.debug('[django-eb-sqs] Connecting to SQS: {}'.format(', '.join(queue_names)))

        sqs = boto3.resource(
            'sqs',
            region_name=settings.AWS_REGION,
            config=Config(retries={'max_attempts': settings.AWS_MAX_RETRIES})
        )

        prefixes = list(filter(lambda qn: qn.startswith(self._PREFIX_STR), queue_names))
        queues = self.get_queues_by_names(sqs, list(set(queue_names) - set(prefixes)))

        queue_prefixes = [prefix.split(self._PREFIX_STR)[1] for prefix in prefixes]
        static_queues = queues
        last_update_time = timezone.now() - timedelta(seconds=settings.REFRESH_PREFIX_QUEUES_S)

        self.write_healthcheck_file()
        last_healthcheck_time = timezone.now()

        logger.debug('[django-eb-sqs] Connected to SQS: {}'.format(', '.join(queue_names)))

        worker = WorkerFactory.default().create()

        logger.info('[django-eb-sqs] WAIT_TIME_S = {}'.format(settings.WAIT_TIME_S))
        logger.info('[django-eb-sqs] NO_QUEUES_WAIT_TIME_S = {}'.format(settings.NO_QUEUES_WAIT_TIME_S))
        logger.info('[django-eb-sqs] MAX_NUMBER_OF_MESSAGES = {}'.format(settings.MAX_NUMBER_OF_MESSAGES))
        logger.info('[django-eb-sqs] AUTO_ADD_QUEUE = {}'.format(settings.AUTO_ADD_QUEUE))
        logger.info('[django-eb-sqs] QUEUE_PREFIX = {}'.format(settings.QUEUE_PREFIX))
        logger.info('[django-eb-sqs] DEFAULT_QUEUE = {}'.format(settings.DEFAULT_QUEUE))
        logger.info('[django-eb-sqs] DEFAULT_MAX_RETRIES = {}'.format(settings.DEFAULT_MAX_RETRIES))
        logger.info('[django-eb-sqs] REFRESH_PREFIX_QUEUES_S = {}'.format(settings.REFRESH_PREFIX_QUEUES_S))

        while True:
            if len(queue_prefixes) > 0 and \
                    timezone.now() - timedelta(seconds=settings.REFRESH_PREFIX_QUEUES_S) > last_update_time:
                queues = static_queues + self.get_queues_by_prefixes(sqs, queue_prefixes)
                last_update_time = timezone.now()
                logger.debug('[django-eb-sqs] Updated SQS queues: {}'.format(
                    ', '.join([queue.url for queue in queues])
                ))

            logger.debug('[django-eb-sqs] Processing {} queues'.format(len(queues)))
            if len(queues) == 0:
                sleep(settings.NO_QUEUES_WAIT_TIME_S)
            else:
                self.process_messages(queues, worker, static_queues)

            if timezone.now() - timedelta(seconds=settings.MIN_HEALTHCHECK_WRITE_PERIOD_S) > last_healthcheck_time:
                self.write_healthcheck_file()
                last_healthcheck_time = timezone.now()

    def process_messages(self, queues, worker, static_queues):
        # type: (list, Worker, list) -> None
        for queue in queues:
            try:
                messages = self.poll_messages(queue)
                logger.debug('[django-eb-sqs] Polled {} messages'.format(len(messages)))

                self._send_signal(MESSAGES_RECEIVED, messages=messages)

                msg_entries = []
                for msg in messages:
                    self._execute_user_code(lambda: self._process_message(msg, worker))
                    msg_entries.append({
                            'Id': msg.message_id,
                            'ReceiptHandle': msg.receipt_handle
                    })

                self._send_signal(MESSAGES_PROCESSED, messages=messages)

                self.delete_messages(queue, msg_entries)

                self._send_signal(MESSAGES_DELETED, messages=messages)
            except ClientError as exc:
                error_code = exc.response.get('Error', {}).get('Code', None)
                if error_code == 'AWS.SimpleQueueService.NonExistentQueue' and queue not in static_queues:
                    logger.debug('[django-eb-sqs] Queue was already deleted {}: {}'.format(queue.url, exc), exc_info=True)
                else:
                    logger.warning('[django-eb-sqs] Error polling queue {}: {}'.format(queue.url, exc), exc_info=True)
            except Exception as exc:
                logger.warning('[django-eb-sqs] Error polling queue {}: {}'.format(queue.url, exc), exc_info=True)

    def delete_messages(self, queue, msg_entries):
        # type: (Queue, list) -> None
        if len(msg_entries) > 0:
            response = queue.delete_messages(Entries=msg_entries)

            # logging
            failed = response.get('Failed', [])
            num_failed = len(failed)
            if num_failed > 0:
                logger.warning('[django-eb-sqs] Failed deleting {} messages: {}'.format(num_failed, failed))

    def poll_messages(self, queue):
        # type: (Queue) -> list
        return queue.receive_messages(
            MaxNumberOfMessages=settings.MAX_NUMBER_OF_MESSAGES,
            WaitTimeSeconds=settings.WAIT_TIME_S,
            AttributeNames=[self._RECEIVE_COUNT_ATTRIBUTE]
        )

    def _send_signal(self, signal, messages):
        # type: (django.dispatch.Signal, list) -> None
        if signal.has_listeners(sender=self.__class__):
            self._execute_user_code(lambda: signal.send(sender=self.__class__, messages=messages))

    def _process_message(self, msg, worker):
        # type: (Message, Worker) -> None
        logger.debug('[django-eb-sqs] Read message {}'.format(msg.message_id))
        try:
            receive_count = int(msg.attributes[self._RECEIVE_COUNT_ATTRIBUTE])

            if receive_count > 1:
                logger.warning('[django-eb-sqs] SQS re-queued message {} times - msg: {}'.format(
                    receive_count, msg.body
                ))

            worker.execute(msg.body)

            logger.debug('[django-eb-sqs] Processed message {}'.format(msg.message_id))
        except ExecutionFailedException as exc:
            logger.warning('[django-eb-sqs] Handling message {} got error: {}'.format(msg.message_id, repr(exc)))

    @staticmethod
    def _execute_user_code(function):
        # type: (Any) -> None
        try:
            with django_db_management():
                function()
        except Exception as exc:
            logger.error('[django-eb-sqs] Unhandled error: {}'.format(exc), exc_info=True)

    def get_queues_by_names(self, sqs, queue_names):
        # type: (ServiceResource, list) -> list
        return [sqs.get_queue_by_name(QueueName=queue_name) for queue_name in queue_names]

    def get_queues_by_prefixes(self, sqs, prefixes):
        # type: (ServiceResource, list) -> list
        queues = []

        for prefix in prefixes:
            queues += sqs.queues.filter(QueueNamePrefix=prefix)

        return queues

    def write_healthcheck_file(self):
        # type: () -> None
        with open(settings.HEALTHCHECK_FILE_NAME, 'w') as file:
            file.write(timezone.now().isoformat())



import os
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))
VERSION = open(os.path.join(here, 'VERSION')).read()
README = open(os.path.join(here, 'README.md')).read()

setup(
    name='django-eb-sqs',
    version=VERSION,
    package_dir={'eb_sqs': 'eb_sqs'},
    include_package_data=True,
    packages=find_packages(),
    description='A simple task manager for AWS SQS',
    long_description=README,
    long_description_content_type="text/markdown",
    url='https://github.com/cuda-networks/django-eb-sqs',
    install_requires=[
        'boto3>=1.9.86',
        'Django>=1.10.6',
        'requests>=2.10.0',
    ],
    classifiers=[
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 2.7',
        'Topic :: Software Development',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Framework :: Django'
    ]
)
